
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdbool.h>

#include "nordic_common.h"
#include "app_error.h"
#include "app_uart.h"

#include "nrf.h"
#include "ble_db_discovery.h"

#include "app_timer.h"
#include "app_trace.h"
#include "app_util.h"
#include "app_error.h"

#include "ble.h"
#include "ble_gap.h"
#include "ble_hci.h"
#include "ble_advdata.h"

#include "nrf_gpio.h"
#include "app_scheduler.h"
#include "softdevice_handler.h"
#include "app_timer_appsh.h"


#include "SEGGER_RTT.h"

#define RB0_PIN													10 // 10 (for SFST-100 Rev3)
#define RB1_PIN													9  //  9 (for SFST-100 Rev3)
#define RG8_PIN													0  //  0 (for SFST-100 Rev3)
#define RG9_PIN													1  //  1 (for SFST-100 Rev3)

#define UART_RX_PIN											RG8_PIN  //Connects to PIC RG8
#define UART_TX_PIN											RG9_PIN  //Connects to PIC RG9

#define UART_TX_BUF_SIZE        				256                             /**< UART TX buffer size. */
#define UART_RX_BUF_SIZE        				256                             /**< UART RX buffer size. */


/**** SPI INTERFACE DEFINES ****/



#define APP_TIMER_PRESCALER             0                                                                                
#define APP_TIMER_OP_QUEUE_SIZE         4                                          

#define IO_TOGGLE_FREQUENCY			        APP_TIMER_TICKS(1000, APP_TIMER_PRESCALER)


APP_TIMER_DEF(m_io_pin_toggle_timer_id);


#define SCHED_MAX_EVENT_DATA_SIZE       sizeof(app_timer_event_t)                 
#define SCHED_QUEUE_SIZE                10            


#define SCAN_INTERVAL           0x00A0    //100ms               /**< Determines scan interval in units of 0.625 millisecond. */
#define SCAN_WINDOW             0x0050    //50ms                /**< Determines scan window in units of 0.625 millisecond. */
#define SCAN_ACTIVE             1                               /**< If 1, performe active scanning (scan requests). */
#define SCAN_SELECTIVE          0                               /**< If 1, ignore unknown devices (non whitelisted). */
#define SCAN_TIMEOUT            0x0000                          /**< */
                                                                
#define MIN_CONNECTION_INTERVAL MSEC_TO_UNITS(20, UNIT_1_25_MS) /**< Determines minimum connection interval in millisecond. */
#define MAX_CONNECTION_INTERVAL MSEC_TO_UNITS(75, UNIT_1_25_MS) /**< Determines maximum connection interval in millisecond. */
#define SLAVE_LATENCY           0                               /**< Determines slave latency in counts of connection events. */
#define SUPERVISION_TIMEOUT     MSEC_TO_UNITS(4000, UNIT_10_MS) /**< Determines supervision time-out in units of 10 millisecond. */
                                                                
#define UUID16_SIZE             2                               /**< Size of 16 bit UUID */
#define UUID32_SIZE	            4                               /**< Size of 32 bit UUID */
#define UUID128_SIZE            16                              /**< Size of 128 bit UUID */

//GAP Advertising packet format: Name (5 bytes) Flags (1 byte) Service ID (10 bytes) + 3 field length bytes + 3 field type bytes = 22 bytes
#define SFTS_ADV_PACKET_DATA_LEN 				22
#define SFTS_ADV_PACKET_DATA_BYTES {    0x06, 0x09, 0x53, 0x46, 0x54, 0x61, 0x67,   \
																				0x02, 0x01, 0x06, 													\
																				0x0B, 0xFF, 0x54, 0x41, 0x48, 0xDB, 0xA6, 0xA8, 0x12, 0x12, 0x16, 0x10 }
//splitfast tag service short ID
#define SPLITFAST_ATHLETE_ID_BASE {0x6c, 0x0e, 0x92, 0xb8, 0x00, 0x00}
#define SFTAG_ADV_TIMEOUT	      10     //if no adv report received for 10 seconds, Tag is removed from found list
#define SFTAG_FOUND_LIST_SIZE   5      //list holds 5 tags found with highest RSSI
#define SFTAG_LOST_SIGNAL_RSSI  -150


typedef struct
{
	bool		  is_active;
	uint8_t   address[6];
	int16_t   rssi;
	uint8_t   timeout;
	
} sf_tag_peripheral_t;


static sf_tag_peripheral_t      m_found_tag_list[SFTAG_FOUND_LIST_SIZE];
static ble_db_discovery_t       m_ble_db_discovery;
uint8_t 												m_num_sf_tags_found = 0;
int16_t 												m_weakest_found_tag_rssi = SFTAG_LOST_SIGNAL_RSSI;
uint8_t													uart_print_char = 0x00;



static uint8_t counter;

//this macro let's us use standard printf() functionality with the SEGGER Windows Serial Terminal via USB
//it requires #define <stdio.h>
#define RTT_PRINTF(...)\
do{\
   char str[64];\
	sprintf(str, __VA_ARGS__);\
	SEGGER_RTT_WriteString(0, str);\
}while(0)



//Connection parameters requested for connection.
static const ble_gap_conn_params_t m_connection_param =
  {
    (uint16_t)MIN_CONNECTION_INTERVAL,   // Minimum connection
    (uint16_t)MAX_CONNECTION_INTERVAL,   // Maximum connection
    0,                                   // Slave latency
    (uint16_t)SUPERVISION_TIMEOUT        // Supervision time-out
  };

	
	
// Parameters used when scanning
static const ble_gap_scan_params_t m_scan_params = 
  {
    .active      = SCAN_ACTIVE,
    .selective   = SCAN_SELECTIVE,
    .p_whitelist = NULL,
    .interval    = SCAN_INTERVAL,
    .window      = SCAN_WINDOW,
    .timeout     = SCAN_TIMEOUT
  };


	
	
	
	
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Callback function for asserts in the SoftDevice.
//
// This function will be called in case of an assert in the SoftDevice
//
// warning: This handler is an example only and does not fit a final product. You need to analyze
// how your product is supposed to react in case of Assert.
//
// warning: On assert from the SoftDevice, the system can only recover on reset.
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void assert_nrf_callback(uint16_t line_num, const uint8_t * p_file_name)
{
    app_error_handler(0xDEADBEEF, line_num, p_file_name);
}









/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Function to clear the list of found SplitFast Tags
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
static void found_tag_list_init(void)
{
	 for(int i=0; i<SFTAG_FOUND_LIST_SIZE; i++)
	{
		m_found_tag_list[i].is_active = false;
		m_found_tag_list[i].rssi = SFTAG_LOST_SIGNAL_RSSI;
		m_found_tag_list[i].timeout = SFTAG_ADV_TIMEOUT;
	}
	m_weakest_found_tag_rssi = SFTAG_LOST_SIGNAL_RSSI;
	m_num_sf_tags_found = 0;
	
}




/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Function called by tags list update timer
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
static void manage_discovered_tags_list(void)
{
   //cycle through the list and manage the timeouts
	if(m_num_sf_tags_found > 0)
	{
			for(int i=0; i<SFTAG_FOUND_LIST_SIZE; i++)
			{
				if(m_found_tag_list[i].is_active)
				{
					if(++m_found_tag_list[i].timeout > SFTAG_ADV_TIMEOUT)
					{
						m_found_tag_list[i].is_active = false;
						m_num_sf_tags_found--;
					}
				}
			}
   }
	 
 }


 
 
static void scan_start(void)
{
    uint32_t err_code;
    found_tag_list_init();
    err_code = sd_ble_gap_scan_start(&m_scan_params);
    APP_ERROR_CHECK(err_code);
}




static bool connect_to_peripheral(ble_gap_addr_t *device_address)
{
   uint32_t err_code;
	
	 err_code = sd_ble_gap_connect(device_address, &m_scan_params, &m_connection_param);
                
   if (err_code == NRF_SUCCESS)
		  return true;
	 else
		 return false;
}



/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//   nRF51 internal UART port Interrupt Service Routine
//
//   Function for handling data received over the nRF51 UART called by Nordic "app_uart.c" module
//
//   This function will receive a single character from the app_uart module and append it to
//   a string. The string will be be sent over BLE when the last character received was a
//   'new line' i.e '\n' (hex 0x0D) or if the string has reached a length of NUS_MAX_DATA_LENGTH
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void uart_event_handle(app_uart_evt_t * p_event)
{
    static uint8_t data_array[16];
	  static uint8_t index = 0;
    
    switch (p_event->evt_type)
    {
        //Handling data from UART
        case APP_UART_DATA_READY:
					 
            UNUSED_VARIABLE(app_uart_get(&data_array[index]));
				
						// do something with the UART data
				
            //index++;
            break;
        //Handling data from UART
        case APP_UART_COMMUNICATION_ERROR:
            APP_ERROR_HANDLER(p_event->data.error_communication);
            break;

        case APP_UART_FIFO_ERROR:
            APP_ERROR_HANDLER(p_event->data.error_code);
            break;

        default:
            break;
						
    }
}




//-----------------------------------------------------------------------
//USED FOR DEBUGGING
//-----------------------------------------------------------------------
static void print_adv_data(const ble_gap_evt_adv_report_t *p_adv_report)
{

	
	
	RTT_PRINTF("Found SFTag Addr:%02X%02X%02X%02X%02X%02X RSSI:%d\n",
                             p_adv_report->peer_addr.addr[5], 
                             p_adv_report->peer_addr.addr[4], 
                             p_adv_report->peer_addr.addr[3], 
                             p_adv_report->peer_addr.addr[2], 
                             p_adv_report->peer_addr.addr[1], 
                             p_adv_report->peer_addr.addr[0],
														 p_adv_report->rssi  
                             );
	//RTT_PRINTF("Addr Type: %u\n", p_adv_report->peer_addr.addr_type);
	//RTT_PRINTF("RSSI: %d\n", p_adv_report->rssi);
	//RTT_PRINTF("Scan Resp: %x\n", p_adv_report->scan_rsp);
	//RTT_PRINTF("Adv Report Type: %x\n", p_adv_report->type);
	//RTT_PRINTF("Data Length: %u\n", p_adv_report->dlen);
	//RTT_PRINTF("Data: ");
	//for(int i=0; i<p_adv_report->dlen; i++)
	//{
	//	RTT_PRINTF("%02X ", p_adv_report->data[i]);
	//}
	//RTT_PRINTF("\n\n");	
}




/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// This function checks the advertising report data received from the peripheral
// and determines if it matches a SplitFast Tag device. If a SplitFast Tag is found 
// it updates the list of found tags by either adding it to the list, or updating
// the record it the data is from a tag we already found, or replacing a weaker signal
// tag in the list if the list is full
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
static void check_if_found_splitfast_tag(const ble_gap_evt_adv_report_t *p_adv_report)
{
	 bool      found_match   = false;
	 bool      new_tag_found = false;
	 int16_t   weakest_rssi;
	 uint8_t   replacement_index;
	
	
	 if(p_adv_report->dlen != SFTS_ADV_PACKET_DATA_LEN) //there are 22 bytes in a tag advertising data packet
		 return;																		//if we get anything else, this is not a SplitFast Tag
		
		uint8_t sf_adv_packet[] = SFTS_ADV_PACKET_DATA_BYTES;  //use this to compare the data in the memcmp()
		
		if(memcmp(sf_adv_packet, p_adv_report->data, SFTS_ADV_PACKET_DATA_LEN) == 0) //check if the advertising data packet matches a tag
		{
			//We found a SplitFast Tag, so add it to the list of found tags
			if(m_num_sf_tags_found == 0) //if there are no tags in the list, then add it right away
			{
				 memcpy(m_found_tag_list[0].address, p_adv_report->peer_addr.addr, 6); 
				 m_found_tag_list[0].rssi = p_adv_report->rssi;
				 m_found_tag_list[0].timeout = 0;
				 m_found_tag_list[0].is_active = true;
				 m_weakest_found_tag_rssi = p_adv_report->rssi;
				 m_num_sf_tags_found++;
				 new_tag_found = true;
			} 
			else //there are already some tags in the list
			{
				  for(int i=0; i<SFTAG_FOUND_LIST_SIZE; i++)  //check the adresses to see if the new tag is already in the list 
				  {
					    if(m_found_tag_list[i].is_active) //only check active tags in the list
					    {
						      if(memcmp(m_found_tag_list[i].address, p_adv_report->peer_addr.addr, 6) == 0) //if the device address matches
						      {
							        //this tag is already in the list, so update it's rssi and reset the timeout counter  
										  found_match = true; 
	  						      m_found_tag_list[i].rssi = p_adv_report->rssi;
		  					      m_found_tag_list[i].timeout = 0;
										  break; //found a match so jump out of the for loop
			  			    }
				  	   }
				  }
					if(!found_match) //the new tag was not in the list, so we need to add it
					{
						  new_tag_found = true;  
						  //if the list isn't full, we add the new tag to the first inactive slot
						  if(m_num_sf_tags_found < SFTAG_FOUND_LIST_SIZE)
							{
								for(int i=0; i<SFTAG_FOUND_LIST_SIZE; i++)
								{
									if(!m_found_tag_list[i].is_active) //found a slot with an inactive tag record
									{
										memcpy(m_found_tag_list[i].address, p_adv_report->peer_addr.addr, 6); 
										m_found_tag_list[i].rssi = p_adv_report->rssi;
										m_found_tag_list[i].timeout = 0;
										m_found_tag_list[i].is_active = true;
										m_num_sf_tags_found++;
									}
								}
							}
							else //else the list is already full, so we need to replace one of the existing tags
							{    //Determine if the new tag has a stonger rssi than the tag in the list with the weakest rssi
								   replacement_index = 0;
								   weakest_rssi = m_found_tag_list[0].rssi;
								
								   for(int i=1; i<SFTAG_FOUND_LIST_SIZE; i++)
								   {
									    if( m_found_tag_list[i].rssi < weakest_rssi)
									    {
										     replacement_index = i; 
										     weakest_rssi = m_found_tag_list[i].rssi;
									    }
							     }
								   //we know the signal and list index of the weakest tag so compare it to the new tag
								   if(p_adv_report->rssi > weakest_rssi) //our new tag has a stronger signal, so replace the weakest tag in the list
								   {
									     memcpy(m_found_tag_list[replacement_index].address, p_adv_report->peer_addr.addr, 6); 
										   m_found_tag_list[replacement_index].rssi = p_adv_report->rssi;
										   m_found_tag_list[replacement_index].timeout = 0;
										   m_found_tag_list[replacement_index].is_active = true;
								   }
					    }
					}//end if(!found_match)
			}//end else there are already tags in the list
			
		}//end of if this is a SplitFast Tag
		if(new_tag_found)
			print_adv_data(p_adv_report);
}





/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// HANDLER FOR BLE EVENTS forwarded from ble_evt_dispatch() in main.c
//
// If we are scanning, this function will be called with Event ID of BLE_GAP_EVT_ADV_REPORT
// when a peripheral is found
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
static void on_ble_evt(ble_evt_t * p_ble_evt)
{
    uint32_t              err_code;
    const ble_gap_evt_t * p_gap_evt = &p_ble_evt->evt.gap_evt;	
    
    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_ADV_REPORT: //Received an advertising packet from a peripheral
        {
  					const ble_gap_evt_adv_report_t *p_adv_report = &p_gap_evt->params.adv_report;
					  if( !p_adv_report->scan_rsp )
						{
						   check_if_found_splitfast_tag(p_adv_report);
            }
	          break;
        }
        
        case BLE_GAP_EVT_CONNECTED:
            RTT_PRINTF("Connected to target\n");
            //m_ble_nus_c.conn_handle = p_ble_evt->evt.gap_evt.conn_handle;
            // start discovery of services. The NUS Client waits for a discovery result
            //err_code = ble_db_discovery_start(&m_ble_db_discovery, p_ble_evt->evt.gap_evt.conn_handle);
            //APP_ERROR_CHECK(err_code);
            break;
    
        case BLE_GAP_EVT_TIMEOUT:
            if (p_gap_evt->params.timeout.src == BLE_GAP_TIMEOUT_SRC_SCAN)
            {
                RTT_PRINTF("[ERROR]: Scan timed out\n");
                scan_start();
            }
            else if (p_gap_evt->params.timeout.src == BLE_GAP_TIMEOUT_SRC_CONN)
            {
                RTT_PRINTF("[ERROR]: Connection Request timed out\n");
            }
            break;
            
        case BLE_GAP_EVT_SEC_PARAMS_REQUEST:
            // Pairing not supported
            err_code = sd_ble_gap_sec_params_reply(p_ble_evt->evt.gap_evt.conn_handle, BLE_GAP_SEC_STATUS_PAIRING_NOT_SUPP, NULL, NULL);
            APP_ERROR_CHECK(err_code);
            break;
    
        case BLE_GAP_EVT_CONN_PARAM_UPDATE_REQUEST:
            // Accepting parameters requested by peripheral
            err_code = sd_ble_gap_conn_param_update(p_gap_evt->conn_handle,
                                                    &p_gap_evt->params.conn_param_update_request.conn_params);
            APP_ERROR_CHECK(err_code);
            break;
    
        default:
            break;
    }
}




/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// ALL BLE EVENTS Generated BY THE SOFTDEVICE ARE SENT HERE
//
// This function is called from the scheduler in the main loop after a BLE stack event has
// been received
//
// The event is then forwarded on to all application modules that need to know about BLE events
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
static void ble_evt_dispatch(ble_evt_t * p_ble_evt)
{
    on_ble_evt(p_ble_evt);  
    ble_db_discovery_on_ble_evt(&m_ble_db_discovery, p_ble_evt);
    //ble_nus_c_on_ble_evt(&m_ble_nus_c,p_ble_evt);
}






//------------------------------------------------------
//timer callback - executes every 1000ms
//------------------------------------------------------
static void io_pin_toggle_timer_handler(void *p_context)
{
	manage_discovered_tags_list();
	counter++;
	
	switch(counter)
	{
		case 1:
			nrf_gpio_pin_set(RB0_PIN);
		break;
		
		case 2:
			nrf_gpio_pin_set(RB1_PIN);
		break;
		
		case 4:
			nrf_gpio_pin_clear(RB0_PIN);
		  nrf_gpio_pin_clear(RB1_PIN);
		break;
		
		case 5:
			counter = 0;
		  app_uart_put( ++uart_print_char );
		break;		
	}
	
}




static void gpio_init(void)
{
	  nrf_gpio_cfg_output(RB0_PIN);
	  nrf_gpio_cfg_output(RB1_PIN);
}


static void timers_create(void)
{
    uint32_t err_code;  
	  
    //APP_TIMER_INIT(APP_TIMER_PRESCALER, APP_TIMER_OP_QUEUE_SIZE, false);
		APP_TIMER_APPSH_INIT(APP_TIMER_PRESCALER, APP_TIMER_OP_QUEUE_SIZE, true);

	  err_code = app_timer_create(&m_io_pin_toggle_timer_id, APP_TIMER_MODE_REPEATED, io_pin_toggle_timer_handler);
    APP_ERROR_CHECK(err_code);

	
}


static void timers_start(void)
{
    uint32_t err_code;
		
	  err_code = app_timer_start(m_io_pin_toggle_timer_id, IO_TOGGLE_FREQUENCY, NULL);
    APP_ERROR_CHECK(err_code); 
}



static void scheduler_init(void)
{
    APP_SCHED_INIT(SCHED_MAX_EVENT_DATA_SIZE, SCHED_QUEUE_SIZE);
}



static void ble_stack_init(void)
{
	  uint32_t err_code;  
	  
	  //Initialize the SoftDevice handler module.
    SOFTDEVICE_HANDLER_INIT(NRF_CLOCK_LFCLKSRC_XTAL_20_PPM, NULL);
	  
    // Enable BLE stack.
    ble_enable_params_t ble_enable_params;
    memset(&ble_enable_params, 0, sizeof(ble_enable_params));
		#ifdef S130
      ble_enable_params.gatts_enable_params.attr_tab_size   = BLE_GATTS_ATTR_TAB_SIZE_DEFAULT;
		#endif
    ble_enable_params.gatts_enable_params.service_changed = false;
		#ifdef S120
      ble_enable_params.gap_enable_params.role              = BLE_GAP_ROLE_CENTRAL;
		#endif
    err_code = sd_ble_enable(&ble_enable_params);
    APP_ERROR_CHECK(err_code);

    // Register with the SoftDevice handler module for BLE events.
    err_code = softdevice_ble_evt_handler_set(ble_evt_dispatch);
    APP_ERROR_CHECK(err_code);
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//  Initialize the nRF51 UART port
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
static void uart_init(void)
{
    uint32_t err_code;

    const app_uart_comm_params_t comm_params =
      {
        .rx_pin_no    = UART_RX_PIN,
        .tx_pin_no    = UART_TX_PIN,
        //.rts_pin_no   = NULL,
        //.cts_pin_no   = NULL,
        .flow_control = APP_UART_FLOW_CONTROL_DISABLED,
        .use_parity   = false,
        .baud_rate    = UART_BAUDRATE_BAUDRATE_Baud38400
      };

    APP_UART_FIFO_INIT(&comm_params,
                        UART_RX_BUF_SIZE,
                        UART_TX_BUF_SIZE,
                        uart_event_handle,
                        APP_IRQ_PRIORITY_LOW,
                        err_code);

    APP_ERROR_CHECK(err_code);
}


static void power_manage(void)
{
    uint32_t err_code = sd_app_evt_wait();
    APP_ERROR_CHECK(err_code);
}


int main(void)
{
    uint32_t err_code;
	
    RTT_PRINTF("SFST100-H BLE Central\n"); 
 
  	gpio_init();
    timers_create();
	  uart_init();			//this causes target to reset
	  ble_stack_init();
	  scheduler_init();
	 
	  err_code = ble_db_discovery_init();
    APP_ERROR_CHECK(err_code);

    timers_start();
		counter = 0;
	
	  scan_start();
    RTT_PRINTF("Scanning...\n"); 
	 
  
  for (;;)
    {
        app_sched_execute();
        power_manage();
    }
}

